package explicit;

import java.util.Arrays;
import java.util.BitSet;
import java.util.List;

import common.functions.primitive.PairPredicateInt;
import common.iterable.IterableBitSet;
import explicit.graphviz.Decorator;
import explicit.graphviz.ShowRewardDecorator;
import explicit.modelviews.EquivalenceRelationInteger;
import explicit.modelviews.MDPDroppedChoicesCached;
import explicit.modelviews.MDPEquiv;
import explicit.modelviews.MDPEquiv.StateChoicePair;
import explicit.rewards.MDPRewards;
import prism.PrismComponent;
import prism.PrismException;

public class ZeroRewardECQuotient
{
	private MDPEquiv quotient;
	private MDPRewards quotientRewards;
	private int numberOfZMECs;

	private static final boolean debug = false;
	
	private ZeroRewardECQuotient(MDPEquiv quotient, MDPRewards quotientRewards, int numberOfZMECs)
	{
		this.quotient = quotient;
		this.quotientRewards = quotientRewards;
		this.numberOfZMECs = numberOfZMECs;
	}

	public MDP getModel()
	{
		return quotient;
	}

	public MDPRewards getRewards()
	{
		return quotientRewards;
	}

	public int getNumberOfZeroRewardMECs()
	{
		return numberOfZMECs;
	}

	public BitSet getNonRepresentativeStates()
	{
		return quotient.getNonRepresentativeStates();
	}

	public void mapResults(double[] soln) {
		for (int s : new IterableBitSet(quotient.getNonRepresentativeStates())) {
			int representative = quotient.mapStateToRestrictedModel(s);
			soln[s] = soln[representative];
		}
	}

	public static ZeroRewardECQuotient getQuotient(PrismComponent parent, MDP mdp, BitSet restrict, MDPRewards rewards) throws PrismException
	{
		PairPredicateInt positiveRewardChoice = (int s, int i) -> {
			if (rewards.getStateReward(s) > 0)
				return true;
			if (rewards.getTransitionReward(s, i) > 0) {
				return true;
			}
			return false;
		};

		// drop positive reward choices
		MDPDroppedChoicesCached zeroRewMDP = new MDPDroppedChoicesCached(mdp, positiveRewardChoice);
		// compute the MECs in the zero-reward sub-MDP
		ECConsumerStore ecs = new ECConsumerStore(parent, zeroRewMDP);
		ECComputer ecComputer = ECComputerDefault.createECComputer(parent, zeroRewMDP, ecs);
		ecComputer.computeMECStates(restrict);

		if (ecs.getMECStates().size() == 0) {
			return null;
		}

		// the equivalence relation on the states
		EquivalenceRelationInteger equiv = new EquivalenceRelationInteger(ecs.getMECStates());

		// we drop zero reward loops on the equivalence classes
		PairPredicateInt zeroRewardECloop = (int s, int i) -> {
			if (positiveRewardChoice.test(s, i)) {
				return false;
			}

			// return true if all successors t of state s for choice i are in the
			// same equivalence class
			boolean rv = mdp.allSuccessorsMatch(s, i, (int t) -> equiv.test(s,t));
			return rv;
		};

		final MDPDroppedChoicesCached droppedZeroRewardLoops = new MDPDroppedChoicesCached(mdp, zeroRewardECloop);
		if (debug)
			droppedZeroRewardLoops.exportToDotFile("zero-mec-loops-dropped.dot");

		BasicModelTransformation<MDP, MDPEquiv> transform = MDPEquiv.transform(droppedZeroRewardLoops, equiv);
		final MDPEquiv quotient = transform.getTransformedModel();

		MDPRewards quotientRewards = new MDPRewards() {
			@Override
			public double getStateReward(int s)
			{
				return rewards.getStateReward(s);
			}

			@Override
			public double getTransitionReward(int s, int i)
			{
				StateChoicePair mapped = quotient.mapToOriginalModel(s, i);
				int mappedChoiceInOriginal = droppedZeroRewardLoops.mapChoiceToOriginalModel(mapped.getState(), mapped.getChoice());
				return rewards.getTransitionReward(mapped.getState(), mappedChoiceInOriginal);
			}

			@Override
			public MDPRewards liftFromModel(Product<? extends Model> product)
			{
				throw new RuntimeException("Not implemented");
			}

			@Override
			public boolean hasTransitionRewards()
			{
				return rewards.hasTransitionRewards();
			}
		};

		if (debug) {
			List<Decorator> decorators = Arrays.asList(new ShowRewardDecorator(quotientRewards));
			quotient.exportToDotFile("zero-mec-quotient.dot", decorators);
		}

		return new ZeroRewardECQuotient(quotient, quotientRewards, ecs.getMECStates().size());
	}

}
